import os
import string
import yaml

from .creator import TestbedCreator


class Yamltemplate(TestbedCreator):
    """ Yamltemplate class (TestbedCreator)

    Creator for the 'yamltemplate' source. Takes in a YAML file containing template strings
    as described in PEP 292, but using % as the delimiter (rather than $). Prompts for input
    for each identifier, substitutes the given values, and outputs the resulting YAML file.
    Can optionally take a list of values from a second YAML file, and prompt the user to
    override them if desired.

    Args:
        template_file ('str'): The path of the input YAML template file.
        value_file ('str') default=None: The path of a YAML file contains key-value pairs
            to be populated in the template.
        noprompt ('boolean') default=False: If specified, the user will not be prompted
            to override the default values from the value_file.
        delimiter ('str') default='%': If specified, uses the provided string as the
            string template delimiter.

    CLI Argument          |  Class Argument
    ---------------------------------------------
    --template-file=value |  template_file=value
    --value-file=value    |  value_file=value
    --noprompt            |  noprompt=True
    --delimiter=value     |  delimiter=True

    pyATS Examples:
        pyats create testbed yamltemplate --template-file=temp.yaml --output=testbed.yaml
        pyats create testbed yamltemplate --template-file=temp.yaml --value-file values.yaml
            --output=testbed.yaml
        pyats create testbed yamltemplate --template-file=temp.yaml --value-file values.yaml
            --output=testbed.yaml --noprompt

    Examples:
        creator = Yamltemplate(template_file="temp.yaml")
        creator.to_testbed_file("testbed.yaml")
        creator.to_testbed_object()

    """

    def _init_arguments(self):
        """ Specifies the arguments for the creator.

        Returns:
            dict: Arguments for the creator.

        """
        return {
            'required': ['template_file'],
            'optional': {
                'value_file': None,
                'noprompt': False,
                'delimiter': '%'
            }
        }

    def _get_info(self, msg, default=''):
        """ Prompts input from user with optional default.

        Args:
            msg ('str'): The prompt message.
            default ('str'): The default value if no input provided.

        Returns:
            str: The user's input, or the default value if no input provided.

        """
        response = ''
        while not response:
            response = input(msg) or default
        return response

    def _generate(self):
        """ Core implementation of how the testbed data is created.

        Returns:
            dict: The intermediate testbed dictionary.

        """
        if not os.path.exists(self._template_file):
            raise FileNotFoundError(f'File does not exist: {self._template_file}')

        if self._noprompt and not self._value_file:
            raise Exception('noprompt option requires a value file to be specified')

        with open(self._template_file, 'r') as f:
            tmpl_str = f.read()

        tmpl = type('CustomTemplate', (string.Template, object), {'delimiter': self._delimiter})(tmpl_str)

        kwargs = {}
        if self._value_file:
            with open(self._value_file, 'r') as f:
                kwargs = yaml.safe_load(f)

        if not self._noprompt:
            # Find all keys from template file
            keys = [s[1] or s[2] for s in tmpl.pattern.findall(tmpl_str) if s[1] or s[2]]
            # Ignore duplicate keys
            for key in list(dict.fromkeys(keys)):
                if key in kwargs:
                    kwargs[key] = self._get_info(f'{key} ({kwargs[key]}): ', default=kwargs[key])
                else:
                    kwargs[key] = self._get_info(f'{key}: ')

        try:
            sub = tmpl.substitute(kwargs)
        except KeyError as e:
            raise Exception(f'No value found for key "{e.args[0]}"')
        clean_yaml = yaml.safe_load(sub)
        return clean_yaml



from unittest import TestCase, main, mock
from pyats.contrib.creators.yamltemplate import Yamltemplate
from pyats.topology import Testbed
import tempfile
import os


class TestYamltemplate(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.tmpl_str = """devices:
  %{device_name}:
    connections:
      cli:
        ip: %mgmt_ip
        protocol: ssh
    credentials:
      default:
        username: %username
        password: %password
    os: %os
"""

        _, cls.template_file = tempfile.mkstemp(text=True)
        with open(cls.template_file, 'w') as file:
            file.write(cls.tmpl_str)

        cls.expected = """devices:
  hostname:
    connections:
      cli:
        ip: 123.123.123.123
        protocol: ssh
    credentials:
      default:
        password: super
        username: admin
    os: iosxe
"""

    @classmethod
    def tearDownClass(cls):
        try:
            os.remove(cls.template_file)
        except OSError:
            pass

    def setUp(self):
        _, self.output_file = tempfile.mkstemp(text=True)
        _, self.values_file = tempfile.mkstemp(text=True)

    def tearDown(self):
        try:
            os.remove(self.output_file)
            os.remove(self.values_file)
        except OSError:
            pass

    @mock.patch('builtins.input')
    def test_interactive(self, input_function):
        input_value = ["hostname", "123.123.123.123", "admin", "super", "iosxe"]
        value = input_value.copy()

        def mock_input(text):
            return value.pop(0)
        input_function.side_effect = mock_input
        Yamltemplate(template_file=self.template_file).to_testbed_file(self.output_file)
        with open(self.output_file) as file:
            self.assertEqual(file.read(), self.expected)
        os.remove(self.output_file)
        value = input_value.copy()
        testbed = Yamltemplate(template_file=self.template_file).to_testbed_object()
        self.assertTrue(isinstance(testbed, Testbed))
        self.assertIn('hostname', testbed.devices)
        self.assertEqual(testbed.devices['hostname'].os, 'iosxe')
        self.assertIn('cli', testbed.devices['hostname'].connections)
        self.assertEqual('123.123.123.123',  testbed.devices['hostname'].connections.cli.ip)
        self.assertEqual('ssh', testbed.devices['hostname'].connections.cli.protocol)
        self.assertIn('default', testbed.devices['hostname'].credentials)
        self.assertEqual('admin', testbed.devices['hostname'].credentials.default.username)

    @mock.patch('builtins.input')
    def test_interactive_with_values_file(self, input_function):
        input_value = ["", "", "admin", "super", "iosxe"]
        value = input_value.copy()

        def mock_input(text):
            return value.pop(0)
        input_function.side_effect = mock_input
        values = """device_name: hostname
mgmt_ip: 123.123.123.123
"""
        with open(self.values_file, 'w') as file:
            file.write(values)
        Yamltemplate(template_file=self.template_file, value_file=self.values_file).to_testbed_file(self.output_file)
        with open(self.output_file) as file:
            self.assertEqual(file.read(), self.expected)
        os.remove(self.output_file)
        value = input_value.copy()
        testbed = Yamltemplate(template_file=self.template_file, value_file=self.values_file).to_testbed_object()
        self.assertTrue(isinstance(testbed, Testbed))
        self.assertIn('hostname', testbed.devices)
        self.assertEqual(testbed.devices['hostname'].os, 'iosxe')
        self.assertIn('cli', testbed.devices['hostname'].connections)
        self.assertEqual('123.123.123.123',  testbed.devices['hostname'].connections.cli.ip)
        self.assertEqual('ssh', testbed.devices['hostname'].connections.cli.protocol)
        self.assertIn('default', testbed.devices['hostname'].credentials)
        self.assertEqual('admin', testbed.devices['hostname'].credentials.default.username)

    def test_values_file_noprompt(self):
        values = """device_name: hostname
mgmt_ip: 123.123.123.123
username: admin
password: super
os: iosxe
"""
        with open(self.values_file, 'w') as file:
            file.write(values)
        Yamltemplate(template_file=self.template_file,
                     value_file=self.values_file,
                     noprompt=True).to_testbed_file(self.output_file)
        with open(self.output_file) as file:
            self.assertEqual(file.read(), self.expected)
        testbed = Yamltemplate(template_file=self.template_file,
                               value_file=self.values_file,
                               noprompt=True).to_testbed_object()
        self.assertTrue(isinstance(testbed, Testbed))
        self.assertIn('hostname', testbed.devices)
        self.assertEqual(testbed.devices['hostname'].os, 'iosxe')
        self.assertIn('cli', testbed.devices['hostname'].connections)
        self.assertEqual('123.123.123.123',  testbed.devices['hostname'].connections.cli.ip)
        self.assertEqual('ssh', testbed.devices['hostname'].connections.cli.protocol)
        self.assertIn('default', testbed.devices['hostname'].credentials)
        self.assertEqual('admin', testbed.devices['hostname'].credentials.default.username)

    def test_missing_keys(self):
        values = """device_name: hostname
mgmt_ip: 123.123.123.123
username: admin
password: super
"""
        with open(self.values_file, 'w') as file:
            file.write(values)
        with self.assertRaises(Exception):
            Yamltemplate(template_file=self.template_file,
                         value_file=self.values_file,
                         noprompt=True)._generate()

    def test_no_values_file_noprompt(self):
        with self.assertRaises(Exception):
            Yamltemplate(template_file=self.template_file, noprompt=True)._generate()

    def test_delimiter(self):
        delimiter = '$'
        _, template_file = tempfile.mkstemp(text=True)
        with open(template_file, 'w') as file:
            file.write(self.tmpl_str.replace('%', delimiter))

        values = """device_name: hostname
mgmt_ip: 123.123.123.123
username: admin
password: super
os: iosxe
"""
        with open(self.values_file, 'w') as file:
            file.write(values)
        Yamltemplate(template_file=template_file,
                     value_file=self.values_file,
                     noprompt=True,
                     delimiter=delimiter).to_testbed_file(self.output_file)
        with open(self.output_file) as file:
            self.assertEqual(file.read(), self.expected)

        try:
            os.remove(template_file)
        except OSError:
            pass


if __name__ == '__main__':
    main()



import os
import shutil
import xlwt

from pyats.contrib.creators.file import File
from unittest import TestCase, main
from pyats.topology import Testbed
from pyats.datastructures import Configuration
from pyats.utils import secret_strings


class TestFile(TestCase):
    # set default pyats configuration
    secret_strings.cfg = Configuration()

    def setUp(self):
        self.csv_file = ("hostname,ip,username,password,protocol,os,"
                         "custom:opt1,custom:opt2\nnx-osv-1,172.25.192.90,admin,admin,"
                         "telnet,nxos,ss1,ss2")
        self.expected = """devices:
  nx-osv-1:
    connections:
      cli:
        ip: 172.25.192.90
        protocol: telnet
    credentials:
      default:
        password: admin
        username: admin
      enable:
        password: admin
    custom:
      opt1: ss1
      opt2: ss2
    os: nxos
    type: nxos
"""
        self.expected_encoded = """devices:
  nx-osv-1:
    connections:
      cli:
        ip: 172.25.192.90
        protocol: telnet
    credentials:
      default:
        password: '%ENC{w5HDncOOw53DoQ==}'
        username: admin
      enable:
        password: '%ENC{w5HDncOOw53DoQ==}'
    custom:
      opt1: ss1
      opt2: ss2
    os: nxos
    type: nxos
"""
        self.test_csv = "/tmp/test.csv"
        self.test_excel = "/tmp/test.xls"
        self.output = "/tmp/testbed.yaml"
        with open(self.test_csv, "w") as csv:
            csv.write(self.csv_file)

        self.csv_file_6 = ("hostname,ip,username,password,protocol,os,"
                           "custom:opt1,custom:opt2\nnx-osv-1,[2001:db8:0400:3200:9020:2:4002:1]:22,admin,admin,"
                           "ssh,nxos,ss1,ss2")
        self.expected_6 = r"""devices:
  nx-osv-1:
    connections:
      cli:
        ip: '[2001:db8:0400:3200:9020:2:4002:1]'
        port: 22
        protocol: ssh
    credentials:
      default:
        password: admin
        username: admin
      enable:
        password: admin
    custom:
      opt1: ss1
      opt2: ss2
    ip: '[2001:db8:0400:3200:9020:2:4002:1]:22'
    os: nxos
    type: nxos
"""
        self.test_csv_6 = "/tmp/test_6.csv"
        self.output_6 = "/tmp/testbed_6.yaml"
        with open(self.test_csv_6, "w") as csv_6:
            csv_6.write(self.csv_file_6)

    def test_no_arguments(self):
        with self.assertRaises(Exception):
            File()

    def test_csv_file(self):
        creator = File(path=self.test_csv)
        creator.to_testbed_file(self.output)
        testbed = creator.to_testbed_object()
        self.assertTrue(os.path.isfile(self.output))
        with open(self.output) as file:
            self.assertEqual(file.read(), self.expected)
        self.assertTrue(isinstance(testbed, Testbed))
        self.assertIn('nx-osv-1', testbed.devices)
        self.assertEqual('ss1', testbed.devices['nx-osv-1'].custom.get('opt1'))
        self.assertEqual('ss2', testbed.devices['nx-osv-1'].custom.get('opt2'))
        self.assertEqual(testbed.devices['nx-osv-1'].os, 'nxos')
        self.assertEqual(testbed.devices['nx-osv-1'].type, 'nxos')
        self.assertIn('cli', testbed.devices['nx-osv-1'].connections)
        self.assertEqual('172.25.192.90',
                         testbed.devices['nx-osv-1'].connections.cli.ip)
        self.assertEqual('telnet',
                         testbed.devices['nx-osv-1'].connections.cli.protocol)
        self.assertIn('default', testbed.devices['nx-osv-1'].credentials)
        self.assertIn('enable', testbed.devices['nx-osv-1'].credentials)
        self.assertEqual('admin',
                         testbed.devices['nx-osv-1'].credentials.default.username)

    def test_csv_file_v6(self):
        creator = File(path=self.test_csv_6)
        creator.to_testbed_file(self.output_6)
        testbed = creator.to_testbed_object()
        self.assertTrue(os.path.isfile(self.output_6))
        with open(self.output_6) as file:
            self.assertEqual(file.read(), self.expected_6)
        self.assertTrue(isinstance(testbed, Testbed))
        self.assertIn('nx-osv-1', testbed.devices)
        self.assertEqual(testbed.devices['nx-osv-1'].os, 'nxos')
        self.assertEqual(testbed.devices['nx-osv-1'].type, 'nxos')
        self.assertIn('cli', testbed.devices['nx-osv-1'].connections)
        self.assertEqual('[2001:db8:0400:3200:9020:2:4002:1]', testbed.devices['nx-osv-1'].connections.cli.ip)
        self.assertEqual(22, testbed.devices['nx-osv-1'].connections.cli.port)
        self.assertEqual('ssh', testbed.devices['nx-osv-1'].connections.cli.protocol)
        self.assertIn('default', testbed.devices['nx-osv-1'].credentials)
        self.assertIn('enable', testbed.devices['nx-osv-1'].credentials)
        self.assertEqual('admin', testbed.devices['nx-osv-1'].credentials.default.username)

    def test_encode_password(self):
        File(path=self.test_csv, encode_password=True).to_testbed_file(
            self.output)
        with open(self.output) as file:
            self.assertEqual(file.read(), self.expected_encoded)

    def test_directory(self):
        directory = '/tmp/sources'
        subdir = '/tmp/sources/subdir'
        outdir = '/tmp/testbeds'
        outsubdir = '/tmp/testbeds/subdir'
        if os.path.isdir(directory):
            shutil.rmtree(directory)
        if os.path.isdir(outdir):
            shutil.rmtree(outdir)
        os.mkdir(directory)
        os.mkdir(subdir)
        for i in range(2):
            with open('{}/{}.csv'.format(directory, i), 'w') as file:
                file.write(self.csv_file)
            with open('{}/{}.csv'.format(subdir, i), 'w') as file:
                file.write(self.csv_file)
        creator = File(path=directory)
        creator.to_testbed_file(outdir)
        self.assertEqual(len(creator.to_testbed_object()), 2)
        self.assertTrue(os.path.isfile('{}/0.yaml'.format(outdir)))
        self.assertTrue(os.path.isfile('{}/1.yaml'.format(outdir)))
        with open('{}/0.yaml'.format(outdir)) as file:
            self.assertEqual(file.read(), self.expected)
        with open('{}/1.yaml'.format(outdir)) as file:
            self.assertEqual(file.read(), self.expected)
        shutil.rmtree(outdir)
        creator = File(path=directory, recurse=True)
        creator.to_testbed_file(outdir)
        self.assertEqual(len(creator.to_testbed_object()), 4)
        self.assertTrue(os.path.isfile('{}/0.yaml'.format(outsubdir)))
        self.assertTrue(os.path.isfile('{}/1.yaml'.format(outsubdir)))
        with open('{}/0.yaml'.format(outsubdir)) as file:
            self.assertEqual(file.read(), self.expected)
        with open('{}/1.yaml'.format(outsubdir)) as file:
            self.assertEqual(file.read(), self.expected)

    def test_excel_load(self):
        wb = xlwt.Workbook()
        ws = wb.add_sheet('testbed')
        for i, k in enumerate([
            'hostname', 'ip', 'username', 'password', 'protocol', 'os',
            'custom:opt1', 'custom:opt2'
        ]):
            ws.write(0, i, k)
        for i, k in enumerate([
            'nx-osv-1', '172.25.192.90', 'admin', 'admin', 'telnet', 'nxos',
            'ss1', 'ss2'
        ]):
            ws.write(1, i, k)
        wb.save(self.test_excel)
        File(path=self.test_excel, encode_password=True).to_testbed_file(
            self.output)
        with open(self.output) as file:
            self.assertEqual(file.read(), self.expected_encoded)


if __name__ == '__main__':
    main()



import xlrd
import xlwt
import xlsxwriter
import csv
import os
import logging

from pyats.topology import Testbed
from .creator import TestbedCreator

class Template(TestbedCreator):
    """ Template class (TestbedCreator)

    Creator for the 'template' source. Generates a CSV or Excel template with 
    the nesscary fields to create a testbed. The template can then be populated
    with device data, and converted to testbeds via the 'file' creator.

    Args:
        add_keys ('list') default=None: Any additional keys that should be added
            to the generated template.
        add_custom_keys ('list') default=None: Any additional custom keys that 
            should be added to the generated template.

    CLI Argument                |  Class Argument
    -----------------------------------------------------------------
    --add-keys k1 k2 ...        |  add_keys=['k1', 'k2', ...]
    --add-custom-keys k1 k2 ... |  add_custom_keys=['k1', 'k2', ...]

    pyATS Examples:
        pyats create testbed template --output=testbed.yaml

    Examples:
        # Create a CSV file template
        creator = Template()
        creator.to_testbed_file("template.csv")

    """

    def _init_arguments(self):
        """ Specifies the arguments for the creator.

        Returns:
            dict: Arguments for the creator.

        """
        self._cli_list_arguments.append('--add-keys')
        self._cli_list_arguments.append('--add-custom-keys')

        return {
            'optional': { 
                'add_keys': None,
                'add_custom_keys': None
            }
        }

    def to_testbed_file(self, output_location):
        """ Saves the template file.

        Args:
            output_location ('str'): Where to save the file.
        
        Returns:
            bool: Indication that the operation is successful or not.
        
        """
        self._output = output_location 
        self._generate()
        return True

    def to_testbed_object(self):
        """ Creates testbed object from the source data.
        
        Returns:
            Testbed: The created testbed.
        
        """
        return Testbed(name="testbed")

    def _generate(self):
        """ Core implementation of how the template is created.

        """
        # If supplied additional keys, add to self.keys
        if self._add_keys:
            self._keys.extend(key.lower() for key in self._add_keys)

        # If supplied custom keys, add converted custom keys
        if self._add_custom_keys:
            self._keys.extend("custom:{}".format(key.lower())
                                            for key in self._add_custom_keys)

        # get file extension
        extension = os.path.splitext(self._output)[-1]
        if extension == '.csv':
            self._write_csv(self._output)
        elif extension == '.xls':
            self._write_xls(self._output)
        elif extension == '.xlsx':
            self._write_xlsx(self._output)
        else:
            raise Exception('File type is not csv or excel')

        # return success result with template filename
        if os.path.isfile(self._output):
            self._result['success'].setdefault('template', self._output)

    def _write_csv(self, output):
        """ Helper for writing keys to CSV.
        
        """
        with open(output, 'w') as f:
            writer = csv.writer(f)
            writer.writerow(self._keys)

    def _write_xls(self, output):
        """ Helper for writing keys to XLS.
        
        """
        wb = xlwt.Workbook()
        ws = wb.add_sheet('testbed')
        for i, k in enumerate(self._keys):
            ws.write(0, i, k)
        wb.save(output)

    def _write_xlsx(self, output):
        """ Helper for writing keys to XLSX.
        
        """
        wb = xlsxwriter.Workbook(output)
        ws = wb.add_worksheet('testbed')
        ws.write_row('A1', self._keys)
        wb.close()



import unittest
from unittest import TestCase, main
from pyats.topology import Testbed
from pyats.datastructures import Configuration
from pyats.utils import secret_strings

def check_ansible_installed():
    try:
        global Ansible
        from pyats.contrib.creators.ansible import Ansible
    except ImportError:
        return True
    return False  

@unittest.skipIf(check_ansible_installed(), 'ansible package is not installed')
class TestAnsible(TestCase):

    maxDiff = None
    # set default pyats configuration
    secret_strings.cfg = Configuration()

    def setUp(self):
        self.inventory = """[all:vars]
ansible_connection=network_cli

[iosxe]
R1_xe ansible_host=172.16.1.228
R2_xr ansible_host=172.16.1.229
R3_nx ansible_host=172.16.1.230

[iosxe:vars]
ansible_become=yes
ansible_become_method=enable
ansible_network_os=ios
ansible_user=admin
ansible_ssh_pass=Cisc0123
ansible_become_pass=Cisc0123
"""
        self.inventory_file = "/tmp/inventory.ini"
        self.output = "/tmp/output"
        with open(self.inventory_file, "w") as file:
            file.write(self.inventory)

    def test_no_arguments(self):
        with self.assertRaises(Exception):
            Ansible()
        with self.assertRaises(Exception):
            Ansible(encode_password=True)

    def test_ansible(self):
        expected = """devices:
  R1_xe:
    alias: R1_xe
    connections:
      cli:
        ip: 172.16.1.228
        protocol: ssh
      enable:
        password: Cisc0123
    credentials:
      default:
        password: Cisc0123
        username: admin
    os: ios
    platform: ios
    type: iosxe
  R2_xr:
    alias: R2_xr
    connections:
      cli:
        ip: 172.16.1.229
        protocol: ssh
      enable:
        password: Cisc0123
    credentials:
      default:
        password: Cisc0123
        username: admin
    os: ios
    platform: ios
    type: iosxe
  R3_nx:
    alias: R3_nx
    connections:
      cli:
        ip: 172.16.1.230
        protocol: ssh
      enable:
        password: Cisc0123
    credentials:
      default:
        password: Cisc0123
        username: admin
    os: ios
    platform: ios
    type: iosxe
"""
        creator = Ansible(inventory_name=self.inventory_file)
        creator.to_testbed_file(self.output)
        with open(self.output) as file:
            self.assertEqual(file.read(), expected)
        testbed = creator.to_testbed_object()
        self.assertTrue(isinstance(testbed, Testbed))
        self.assertIn('R1_xe', testbed.devices)
        self.assertIn('R2_xr', testbed.devices)
        self.assertIn('R3_nx', testbed.devices)
        self.assertEqual(testbed.devices['R1_xe'].os, 'ios')
        self.assertEqual(testbed.devices['R1_xe'].type, 'iosxe')
        self.assertEqual(testbed.devices['R1_xe'].platform, 'ios')
        self.assertEqual(testbed.devices['R2_xr'].os, 'ios')
        self.assertEqual(testbed.devices['R2_xr'].type, 'iosxe')
        self.assertEqual(testbed.devices['R2_xr'].platform, 'ios')
        self.assertEqual(testbed.devices['R3_nx'].os, 'ios')
        self.assertEqual(testbed.devices['R3_nx'].type, 'iosxe')
        self.assertEqual(testbed.devices['R3_nx'].platform, 'ios')
        self.assertIn('cli', testbed.devices['R1_xe'].connections)
        self.assertIn('cli', testbed.devices['R2_xr'].connections)
        self.assertIn('cli', testbed.devices['R3_nx'].connections)
        self.assertEqual('172.16.1.228',
                                    testbed.devices['R1_xe'].connections.cli.ip)
        self.assertEqual('172.16.1.229',
                                    testbed.devices['R2_xr'].connections.cli.ip)
        self.assertEqual('172.16.1.230',
                                    testbed.devices['R3_nx'].connections.cli.ip)
        self.assertEqual('ssh', 
                            testbed.devices['R1_xe'].connections.cli.protocol)
        self.assertEqual('ssh', 
                            testbed.devices['R2_xr'].connections.cli.protocol)
        self.assertEqual('ssh', 
                            testbed.devices['R3_nx'].connections.cli.protocol)
        self.assertIn('default', testbed.devices['R1_xe'].credentials)
        self.assertEqual('admin', 
                        testbed.devices['R1_xe'].credentials.default.username)
        self.assertIn('default', testbed.devices['R2_xr'].credentials)
        self.assertEqual('admin', 
                        testbed.devices['R2_xr'].credentials.default.username)
        self.assertIn('default', testbed.devices['R3_nx'].credentials)
        self.assertEqual('admin', 
                        testbed.devices['R3_nx'].credentials.default.username)

    def test_encode_password(self):
        expected = """devices:
  R1_xe:
    alias: R1_xe
    connections:
      cli:
        ip: 172.16.1.228
        protocol: ssh
      enable:
        password: '%ENC{wrPDosOUw5fCo8KQwpbCmA==}'
    credentials:
      default:
        password: '%ENC{wrPDosOUw5fCo8KQwpbCmA==}'
        username: admin
    os: ios
    platform: ios
    type: iosxe
  R2_xr:
    alias: R2_xr
    connections:
      cli:
        ip: 172.16.1.229
        protocol: ssh
      enable:
        password: '%ENC{wrPDosOUw5fCo8KQwpbCmA==}'
    credentials:
      default:
        password: '%ENC{wrPDosOUw5fCo8KQwpbCmA==}'
        username: admin
    os: ios
    platform: ios
    type: iosxe
  R3_nx:
    alias: R3_nx
    connections:
      cli:
        ip: 172.16.1.230
        protocol: ssh
      enable:
        password: '%ENC{wrPDosOUw5fCo8KQwpbCmA==}'
    credentials:
      default:
        password: '%ENC{wrPDosOUw5fCo8KQwpbCmA==}'
        username: admin
    os: ios
    platform: ios
    type: iosxe
""" 
        creator = Ansible(inventory_name=self.inventory_file, 
                                                        encode_password=True)
        creator.to_testbed_file(self.output)
        with open(self.output) as file:
            self.assertEqual(file.read(), expected)

if __name__ == '__main__':
    main()



import requests 
import copy
import logging

from .creator import TestbedCreator

logger = logging.getLogger(__name__)

class Netbox(TestbedCreator):
    """ Netbox class (TestbedCreator)

    Creator for the 'netbox' source. Retrieves device data from a hosted Netbox
    instance via REST API and converts them to either a testbed file or testbed
    object. Will prompt user for device credentials.
    
    The platform_map allows you to take whatever you have in Netbox for Platform
    and map it to a particular platform and os setting in PyATS. If it is not provided
    or matched, it follows the normal processing for platform and os determination.

    The custom_data_source will lookup whatever is provided in the list recursively
    using the self._get_info function. That value will be set as the custom key
    in the testbed for the device.

    Args:
        netbox_url ('str'): The URL to the Netbox instance.
        user_token ('str'): The REST API access token. Can be found under your 
            profile and in the API Tokens tab.
        encode_password (bool) default=False: Should generated testbed encode 
            its passwords.
        topology (bool) default=False: Do not generate topology data by default, 
        verify (bool) default=True: Should requests library validate SSL cert for netbox
        url_filter ('str') default=None: Netbox URL filter string, example: 'status=active&site=test_site'
        def_user ('str') default=None: Set the username for all devices
        def_pass ('str') default=None: Set the password for all devices
        def_enable ('str) default=None: Set the enable password for all devices
        host_upper (bool) default=False: Store hostname in upper case (to match the prompt)
        platform_map (dict) default=None: map Netbox platform to PyATS platform
        custom_data_source (list) default=None: determine source of PyATS custom key

    CLI Argument        |  Class Argument
    ---------------------------------------------
    --netbox-url=value  |  netbox_url=value
    --user-token=value  |  user_token=value
    --encode-password   |  encode_password=True
    --topology          |  topology=True
    --verify=False      |  verify=False
    --host_upper=True   |  host_upper=True
    --url_filter=value  |  url_filter=value
    --def_user=value    |  def_user=value
    --def_pass=value    |  def_pass=value
    --def_enable=value  |  def_enable=value
    --tag_telnet=value  | tag_telnet=value
    --platform_map=value| platform_map=value
    --custom_data_source=value
                        | custom_data_source=value

    pyATS Examples:
        pyats create testbed netbox --output=out --netbox-url=https://netbox.com
        --user-token=72830d67beff4ae178b94d8f781842408df8069d

    Examples:
        # Create testbed from Netbox source
        creator = Netbox(user_token="72830d67", netbox_url="https://netbox.com", 
            platform_map = {
                "dnac": {
                    "platform": "switch",
                    "os": "iosxe"
                },
                "vmanage": {
                    "platform": "sdwan",
                    "os": "iosxe"
                }
            },
            custom_data_source = [
                "config_context",
                "pyats_custom"
            ]
        })
        creator.to_testbed_file("testbed.yaml")
        creator.to_testbed_object()

    """

    def _init_arguments(self):
        """ Specifies the arguments for the creator.

        Returns:
            dict: Arguments for the creator.

        """
        return {
            'required': ['netbox_url', 'user_token'],
            'optional': {
                'encode_password': False,
                'topology': False,
                'verify': True,
                'host_upper': False,
                'url_filter': None,
                'def_user': None,
                'def_pass': None,
                'def_enable': None,
                'tag_telnet': None,
                'platform_map': None,
                'custom_data_source': None
            }
        }

    def _parse_response(self, response, return_property):
        """ Helper to parse JSON response from HTTP requests.

        Args:
            response ('response'): The response object obtained after 
                HTTP request.
            return_property ('str'): Any filtering that will be applied after 
                parsing the response.

        Returns:
            dict: The response JSON in dictionary form or none if response 
                is invalid.

        """
        response = None if not response else response.json()

        if response and return_property:
            return response[return_property]

        return response

    def _get_request(self, url, headers=None, return_property=None):
        """ Helper to send GET request and returns the response JSON in 
            dictionary form.

        Args:
            url ('str'): URL of where to send the GET request to.
            headers ('dict'): The headers used in the HTTP request.
            return_property ('str'): Any filtering applied to the dictionary
                after parsing the JSON.

        Returns:
            dict: The response JSON in dictionary form.
    
        """
        try:
            response = None

            if not headers:
                response = requests.get(url, verify=self._verify)
            else:
                response = requests.get(url, headers=headers, verify=self._verify)
            
            results = self._parse_response(response, return_property)

            while "next" in response.json().keys() and response.json()["next"]:
                next_url = response.json()["next"]
                if not headers:
                    response = requests.get(next_url, verify=self._verify)
                else:
                    response = requests.get(next_url, headers=headers, verify=self._verify)
                
                results += self._parse_response(response, return_property)

            return results
        except:
            return None

    def _format_url(self, base, route):
        """ Helper to join the base URL and its route.

        Args:
            base ('str'): The base of the URL.
            route ('dict'): The URL route.

        Returns:
            str: The joined URL.
        
        """
        return "{}{}{}".format(base, "" if base[-1] == "/" else "/", route)

    def _set_value_if_exists(self, container, key, entry):
        """ Helper to set value in dictionary if given entry
            is not none.

        Args:
            container ('dict'): Where to insert the new key value pair. 
            key ('str'): The key that the entry corresponds to.
            entry ('str'): The value to be added.

        Returns:
            bool: Whether or not the entry is inserted.
        
        """
        if not entry is None:
            container.setdefault(key, entry)

            return True
        
        return False

    def _parse_os(self, os):
        """ Helper to parse the OS type.

        Args:
            os ('str'): The input os type.

        Returns:
            str: The OS type if it exists or None if it cannot
                be found in list of valid OS.
    
        """
        valid_os = ["com", "asa", "dnac", "ios-xe", "ios-xr",
            "iosxe", "iosxr", "ios", "junos", "linux", "nxos", "nx-os", "yang", 
            "ftd"]

        for valid in valid_os:
            if os and valid in os.lower():
                #ftd is a deprecated platform in pyats, change to fxos
                if valid == "ftd":
                   return "fxos"
                return valid.replace("-", "")
        
        return None

    def _format_type(self, interface_name, interface_type): 
        """ Helper to parse the interface type.

        Args:
            interface_type ('str'): The input interface type.
            interface_name ('str'): The input interface name from Netbox.
            interface_type ('str'): The input interface type from Netbox.

        Returns:
            str: The interface type if it exists or none if it 
            cannot be found in list of valid types.
    
        """
        # for reference: available at /api/dcim/_choices/interface:type
        netbox_interface_types = [
            {
                "value": 0,
                "label": "Virtual"
            },
            {
                "value": 200,
                "label": "Link Aggregation Group (LAG)"
            },
            {
                "value": 800,
                "label": "100BASE-TX (10/100ME)"
            },
            {
                "value": 1000,
                "label": "1000BASE-T (1GE)"
            },
            {
                "value": 1120,
                "label": "2.5GBASE-T (2.5GE)"
            },
            {
                "value": 1130,
                "label": "5GBASE-T (5GE)"
            },
            {
                "value": 1150,
                "label": "10GBASE-T (10GE)"
            },
            {
                "value": 1170,
                "label": "10GBASE-CX4 (10GE)"
            },
            {
                "value": 1050,
                "label": "GBIC (1GE)"
            },
            {
                "value": 1100,
                "label": "SFP (1GE)"
            },
            {
                "value": 1200,
                "label": "SFP+ (10GE)"
            },
            {
                "value": 1300,
                "label": "XFP (10GE)"
            },
            {
                "value": 1310,
                "label": "XENPAK (10GE)"
            },
            {
                "value": 1320,
                "label": "X2 (10GE)"
            },
            {
                "value": 1350,
                "label": "SFP28 (25GE)"
            },
            {
                "value": 1400,
                "label": "QSFP+ (40GE)"
            },
            {
                "value": 1420,
                "label": "QSFP28 (50GE)"
            },
            {
                "value": 1500,
                "label": "CFP (100GE)"
            },
            {
                "value": 1510,
                "label": "CFP2 (100GE)"
            },
            {
                "value": 1650,
                "label": "CFP2 (200GE)"
            },
            {
                "value": 1520,
                "label": "CFP4 (100GE)"
            },
            {
                "value": 1550,
                "label": "Cisco CPAK (100GE)"
            },
            {
                "value": 1600,
                "label": "QSFP28 (100GE)"
            },
            {
                "value": 1700,
                "label": "QSFP56 (200GE)"
            },
            {
                "value": 1750,
                "label": "QSFP-DD (400GE)"
            },
            {
                "value": 2600,
                "label": "IEEE 802.11a"
            },
            {
                "value": 2610,
                "label": "IEEE 802.11b/g"
            },
            {
                "value": 2620,
                "label": "IEEE 802.11n"
            },
            {
                "value": 2630,
                "label": "IEEE 802.11ac"
            },
            {
                "value": 2640,
                "label": "IEEE 802.11ad"
            },
            {
                "value": 2810,
                "label": "GSM"
            },
            {
                "value": 2820,
                "label": "CDMA"
            },
            {
                "value": 2830,
                "label": "LTE"
            },
            {
                "value": 6100,
                "label": "OC-3/STM-1"
            },
            {
                "value": 6200,
                "label": "OC-12/STM-4"
            },
            {
                "value": 6300,
                "label": "OC-48/STM-16"
            },
            {
                "value": 6400,
                "label": "OC-192/STM-64"
            },
            {
                "value": 6500,
                "label": "OC-768/STM-256"
            },
            {
                "value": 6600,
                "label": "OC-1920/STM-640"
            },
            {
                "value": 6700,
                "label": "OC-3840/STM-1234"
            },
            {
                "value": 3010,
                "label": "SFP (1GFC)"
            },
            {
                "value": 3020,
                "label": "SFP (2GFC)"
            },
            {
                "value": 3040,
                "label": "SFP (4GFC)"
            },
            {
                "value": 3080,
                "label": "SFP+ (8GFC)"
            },
            {
                "value": 3160,
                "label": "SFP+ (16GFC)"
            },
            {
                "value": 3320,
                "label": "SFP28 (32GFC)"
            },
            {
                "value": 3400,
                "label": "QSFP28 (128GFC)"
            },
            {
                "value": 4000,
                "label": "T1 (1.544 Mbps)"
            },
            {
                "value": 4010,
                "label": "E1 (2.048 Mbps)"
            },
            {
                "value": 4040,
                "label": "T3 (45 Mbps)"
            },
            {
                "value": 4050,
                "label": "E3 (34 Mbps)"
            },
            {
                "value": 5000,
                "label": "Cisco StackWise"
            },
            {
                "value": 5050,
                "label": "Cisco StackWise Plus"
            },
            {
                "value": 5100,
                "label": "Cisco FlexStack"
            },
            {
                "value": 5150,
                "label": "Cisco FlexStack Plus"
            },
            {
                "value": 5200,
                "label": "Juniper VCP"
            },
            {
                "value": 5300,
                "label": "Extreme SummitStack"
            },
            {
                "value": 5310,
                "label": "Extreme SummitStack-128"
            },
            {
                "value": 5320,
                "label": "Extreme SummitStack-256"
            },
            {
                "value": 5330,
                "label": "Extreme SummitStack-512"
            },
            {
                "value": 32767,
                "label": "Other"
            }
        ]

        # map the values of interface type from Netbox to a pyATS/Genie valid type 
        valid_types_lookup = {
            "ethernet": [800, 1000, 1120, 1130, 1150, 1170, 1050, 1100, 1200, 1300, 
                1310, 1320, 1350, 1400, 1420, 1500, 1510, 1650, 1520, 1550, 1600, 1700, 
                1750,], 
            # "loopback": [], 
            "vlan": [0], 
            "port-channel": [200], 
            "wireless": [2600, 2610, 2620, 2630, 2640,], 
            "cellular": [2810, 2820, 2830], 
            "SONET": [6100, 6200, 6300, 6400, 6500, 6600, 6700], 
            "fibrechannel": [3010, 3020, 3040, 3080, 3160, 3320, 3400], 
            "serial": [4000, 4010, 4040, 4050], 
            "stacking": [5000, 5050, 5100, 5150, 5200, 5300, 5310, 5320, 5330], 
            "other": [32767]
            }

        # TODO: iosxr interface-types require UPPER case names - need to update to support 
        valid_types = ["ethernet", "loopback", "vlan", "port-channel", "pseudowire", "tunnel", "mgmt", "nve", ]

        # 2 phase type lookup, first try with interface name, then use Netbox interface type
        for valid in valid_types:
            if interface_name and valid in interface_name:
                return valid

        for valid, netbox_type_values in valid_types_lookup.items(): 
            if interface_type["value"] in netbox_type_values: 
                return valid
        
        # TODO: ASAv Management0/0 interfaces don't match interface name based types, and are "Virtual" interfaces in NetBox

        return None

    def _get_info(self, data, keys, transformation=None):
        """ Helper for getting data from nested dictionary. 

        Args:
            data ('dict'): The dictionary where you want to 
                retrieve info from.
            keys ('list'): An ordered list of keys depicting 
                the nested structure.
            transformation ('callable'): Any transformation 
                to apply to the result data.
        
        Returns:
            str: The retrieved data or None if keys does not
                exist in dictionary.
    
        """
        current = data

        if not keys:
            return None

        for key in keys:    
            if not current or key not in current.keys():
                return None
            
            current = current[key]

        if transformation and current:
            current = transformation(current)

        return current

    def _generate(self):
        """ Transforms NetBox data into testbed format.
        
        Returns:
            dict: The intermediate dictionary format of the testbed data.
    
        """
        logger.info("Begin retrieving data from netbox...")
        token = "Token {}".format(self._user_token)
        headers = { "Authorization": token }
        testbed = {}
        data = {}
        topology = {}

        # Configure Testbed wide details
        if self._def_user and self._def_pass:
            logger.info("Configuring testbed default credentials.")
            testbed["credentials"] = {
                "default": {
                    "username": self._def_user, 
                    "password": self._def_pass
                }
            }
            # Add Enable if provided
            if self._def_enable:
                logger.info("Configuring testbed default enable password.")
                testbed["credentials"]["enable"] = {
                    "password": self._def_enable
                }
            
        # If 'self._verify' is of type 'String', then change it from 'String' to type 'Boolean'
        if isinstance(self._verify, str):
            self._verify = True if self._verify == "True" else False

        response = [] 
        netbox_endpoints = ["dcim/devices", "virtualization/virtual-machines"]
        for endpoint in netbox_endpoints: 
            if self._url_filter is None:
                url="api/{endpoint}/?format=json".format(endpoint=endpoint)
            else:
                url="api/{endpoint}/?format=json&{url_filter}".format(endpoint=endpoint, url_filter=self._url_filter)

            devices_url = self._format_url(self._netbox_url, url)
            response += self._get_request(devices_url, headers, "results")

        # If no response is received for retrieving a list of devices, stop
        if not response: 
            logger.error("\nnetbox instance gave no response")
            return None

        for device in response:
            is_valid = True

            if self._host_upper is True:
                device_name = device["name"].upper()
            else:
                device_name = device["name"]

            logger.info("Retrieving associated data for {}..."
                                                        .format(device_name))
            device_id = device["id"]
            device_data = data.setdefault(device_name, {})

            # If provided, use Platform Map to translate Netbox Platform to PyATS Platform/OS
            platform_map = self._platform_map
            platform_map_success = False
            if platform_map:
                raw_device_platform = self._get_info(device, 
                                ["platform", "slug"], lambda slug: slug.lower())
                if raw_device_platform in platform_map.keys():
                    # Mapping found in platform_map, set platform and os.
                    device_platform = platform_map[raw_device_platform]["platform"]
                    is_valid &= self._set_value_if_exists(device_data, "os",
                                        platform_map[raw_device_platform]["os"])
                    platform_map_success = True

            # If platform mapping was not provided or was unsuccessful, then default to standard behavior for assigning os and platform
            if platform_map_success == False:
                # Construct device platform data
                device_platform = self._parse_os(self._get_info(device, 
                                ["platform", "slug"], lambda slug: slug.lower()))
                # OS value is required and must exist
                is_valid &= self._set_value_if_exists(device_data, "os",
                                                                device_platform)

            # Set other testbed values if they exists
            self._set_value_if_exists(device_data, "alias", device_name)
            self._set_value_if_exists(device_data, "platform", device_platform)
            self._set_value_if_exists(device_data, "type", 
                            self._get_info(device, ["device_type", "model"]))
            self._set_value_if_exists(device_data, "custom",
                            self._get_info(device,self._custom_data_source))

            # NetBox Virtual Machines don't have a "device_type" attribute, but pyATS requires
            # one. If missing, construct a type from Platform + Role
            if "type" not in device_data.keys(): 
                role_name = self._get_info(device, ["role", "name"])
                platform_name = self._get_info(device, ["platform", "name"])
                device_data["type"] = "{platform} - {role}".format(platform = platform_name, role = role_name)
            
            # Initialize connection data
            connections = device_data.setdefault("connections", {})
            cli = connections.setdefault("cli", {})
            mask_filter = lambda ip: ip.split("/")[0]
            ipv6 = self._get_info(device, [
                "primary_ip6", "address"
            ], mask_filter)
            found_ip = False

            if self._tag_telnet is not None and self._tag_telnet in device['tags']:
                protocol='telnet'
            else:
                protocol='ssh'

            cli.setdefault("protocol", protocol)
            # Attempt to set connection protocol to primary IP, if found
            found_ip |= self._set_value_if_exists(cli, "ip", self._get_info(
                            device, ["primary_ip4", "address"], mask_filter))
            found_ip |= self._set_value_if_exists(cli, "ip",
                self._get_info(device, ["primary_ip", "address"], mask_filter))
            found_ip |= self._set_value_if_exists(cli, "ip", ipv6)

            # If we did not find a valid OS type for device, we skip it
            if not is_valid:
                logger.warning(
                    "OS type is not valid for {}. ".format(device_name) +
                    "Skipping..."
                )
                
                # Delete the device from testbed
                del data[device_name]
                continue
            
            # Set IP to IPV6 if IPV4 primary does not exist
            if "ip" in cli and "." in cli["ip"] and ipv6:
                connections.setdefault("ipv6", {
                    "ip": ipv6, "protocol": "ssh"
                })
            
            if self._topology is True:
                # Need to determine whether to do the lookup for interfaces against DCIM or VM
                if "rack" in device.keys():
                    # Even unracked devices have the key "rack" in the returned body
                    interface_url = self._format_url(self._netbox_url, 
                        "api/dcim/interfaces/?device_id={}&format=json".format(
                                                                            device_id))                    
                else: 
                    # Otherwise lookup as VM
                    interface_url = self._format_url(self._netbox_url, 
                        "api/virtualization/interfaces/?virtual_machine_id={}&format=json".format(
                                                                            device_id))                    

                # Send request for interfaces
                interface_response = self._get_request(interface_url, headers, 
                                                                        "results")

                # If no interface response are received, we skip the device
                if not interface_response:
                    logger.warning(
                        "No interface found for {}. ".format(device_name) +
                        "Skipping device..."
                    )

                    # Delete device data from testbed
                    del data[device_name]
                    continue
                
                # Initialize interface data
                interfaces = topology.setdefault(device_name, {
                    "interfaces": {} 
                })["interfaces"]

                for interface in interface_response:
                    interface_name = interface["name"]
                    interface_id = interface["id"]
                    current = interfaces.setdefault(interface_name, {})

                    current.setdefault("alias", "{}_{}"
                                            .format(device_name, interface_name))
                    self._set_value_if_exists(current, "type", 
                                        self._format_type(
                                            interface_name.lower(), 
                                            interface["type"]
                                        ))
                    if current.get('type') is None:
                        logger.info("{device_name} interface {interface_name} is not valid, skipping".format(device_name=device_name, interface_name=interface_name.lower()))
                        del interfaces[interface_name]
                        continue
                    
                    # Use the cable information from Netbox to configure link on interface
                    self._set_value_if_exists(
                        current, "link", 
                        self._get_info(interface, ["cable", "id"], lambda link: "cable_num_{link}".format(link=link))
                    )

                    # Attempt to retrieve IP for each interface
                    ip_url = self._format_url(self._netbox_url,
                        "api/ipam/ip-addresses/?interface_id={}&format=json"
                                                            .format(interface_id))
                    ip_response = self._get_request(ip_url, headers, "results")

                    # If no response for IP retrieval then we skip this interface
                    if not ip_response:
                        continue

                    for ip_address in ip_response:
                        # If we have not found primary IP for the device, set it to
                        # the first interface IP we see
                        if not found_ip:
                            cli.setdefault("ip", mask_filter(ip_address["address"]))

                            found_ip = True
                        
                        # Correctly set IP information on interface data
                        family = "ipv4" if "." in ip_address["address"] else "ipv6"
                        current.setdefault(family, ip_address["address"])

            # If no primary IP found for this device, then we stop and skip
            if not found_ip:
                logger.warning(
                    "Connection IP not found for {}. ".format(device_name) +
                    "Skipping device..."
                )
                
                data.pop(device_name, None)
                topology.pop(device_name, None)
                continue
                
            if self._def_user is None or self._def_pass is None:
                # Request user to manually enter their credentials for each device
                logger.info("Connection Credentials for {}:".format(device_name))

                username = input("Username: ")
                password = input("Password: ")

                # Only include device specific credentials block if default credentials were NOT provided
                device_data.setdefault("credentials", {
                    "default": { "username": username, "password": password }
                })
            else:
                username = self._def_user
                password  = self._def_pass
                device_data.setdefault("credentials", {
                    "default": { "username": username, "password": password }
                })


        # If testbed has data, return it
        if len(data.keys()) > 0:
            return { "testbed": testbed, "devices": data, "topology": topology }
        
        return None



import xlrd

from unittest import TestCase, main
from pyats.topology import Testbed
from pyats.contrib.creators.template import Template


class TestTemplate(TestCase):
    test_csv = '/tmp/test.csv'
    test_excel = '/tmp/test.xls'

    def test_csv_template(self):
        expected = "hostname,ip,username,password,protocol,os\n"
        Template().to_testbed_file(self.test_csv)
        with open(self.test_csv) as file:
            self.assertEqual(file.read(), expected)

    def test_excel_template(self):
        Template().to_testbed_file(self.test_excel)
        expected = ['hostname', 'ip', 'username', 'password', 'protocol', 'os']
        ws = xlrd.open_workbook(self.test_excel).sheet_by_index(0)
        keys = ws.row_values(0)
        self.assertEqual(expected, keys)
        
    def test_to_testbed_object(self):
        self.assertTrue(isinstance(Template().to_testbed_object(), Testbed))

    def test_add_keys(self):
        Template(add_keys=['a', 'b' ,'c']).to_testbed_file(self.test_csv)
        expected = "hostname,ip,username,password,protocol,os,a,b,c\n"
        with open(self.test_csv) as file:
            self.assertEqual(file.read(), expected)
        Template(add_custom_keys=['a', 'b' ,'c']).to_testbed_file(self.test_csv)
        expected = ("hostname,ip,username,password,protocol,os,custom:a,"
                    "custom:b,custom:c\n")
        with open(self.test_csv) as file:
            self.assertEqual(file.read(), expected)

if __name__ == '__main__':
    main()



import logging
import argparse
import ipaddress
from concurrent.futures import ThreadPoolExecutor
from genie.conf.base import Testbed, Device, Interface, Link
from pyats.async_ import pcall
from pyats.log import TaskLogHandler
from pyats.log import ScreenHandler

log = logging.getLogger(__name__)

class TestbedManager(object):
    '''Class designed to handle device interactions for connecting devices
       and cdp and lldp
    '''
    def __init__(self, testbed, supported_os, config=False, ssh_only=False, alias_dict={},
                 timeout=10, logfile = '', disable_config=False):

        self.config = config
        self.ssh_only = ssh_only
        self.testbed = testbed
        self.alias_dict = alias_dict
        self.timeout = int(timeout)
        self.cdp_configured = set()
        self.lldp_configured = set()
        self.visited_devices = set()
        self.supported_os = supported_os
        self.logfile = logfile
        if disable_config:
            self.disable_config = []
        else:
            self.disable_config = None

    def connect_all_devices(self, limit):
        '''Creates a ThreadPoolExecutor designed to connect to each device in parallel
        after it takes the connection results of the objects and sorts them into three
        sets for logging purposes

        Args:
            limit ('int'): max number of threads to spawn
            
        Returns:
            three sets for devices that were connected, failed to connect to, and skipped
        '''
        
        results = {}
        success = set()
        fail = set()
        skip = set()
        
        # Set up a thread pool executor to connect to all devices at the same time
        with ThreadPoolExecutor(max_workers = limit) as executor:
            for device_name, device_obj in self.testbed.devices.items():
                # If already connected or device has already been visited skip
                if device_obj.connected or device_name in self.visited_devices:
                    continue
                if device_obj.os not in self.supported_os:
                    log.debug('     Device {} does not have valid os, skipping'.format(device_name))
                    skip.add(device_name)
                    continue
                log.debug('     Attempting to connect to {device}'.format(device=device_name))
                results[device_name] = executor.submit(self._connect_one_device,
                                                       device_name)
        
        for name, exe in results.items():
            if exe.result():
                success.add(name)
            else:
                fail.add(name)

        return success, fail, skip
        
        
    def _connect_one_device(self, device):
        '''Connect to the given device in the testbed using the given
        connections and after that enable cdp and lldp if allowed

        Args:
            device ('str'): name of device being connected
        '''
        # when -v is used root logger effective level will be set to debug so
        # script will set st_out for devices to true so device data will be sent ton
        # console
        if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
            to_stdout = True
        else:
            to_stdout = False
            
        # if there is a preferred alias for the device, attempt to connect with device
        # using that alias, if the attempt fails or the alias doesn't exist, it will
        # attempt to connect with the default
        if device in self.alias_dict:
            if self.alias_dict[device] in self.testbed.devices[device].connections:
                log.debug('     Attempting to connect to {} with alias {}'.format(device, self.alias_dict[device]))
                try:
                    self.testbed.devices[device].connect(via = str(self.alias_dict[device]),
                                                         connection_timeout=self.timeout,
                                                         log_stdout=to_stdout,
                                                         logfile = self.logfile,
                                                         learn_os = True,
                                                         init_config_commands = self.disable_config)
                    log.debug('     Connected to device {}'.format(device))
                except Exception as e:
                    log.debug('     Failed to connect to {} with alias {}'.format(device, self.alias_dict[device]))
                    self.testbed.devices[device].destroy(str(self.alias_dict[device]))
                else:
                    
                    # No exception raised - get out
                    return self.testbed.devices[device].connected
            else:
                log.debug('     Device {} does not have a connection with alias {}'.format(device, self.alias_dict[device]))

        # Use default - Go through all connection on the device
        for one_connect in self.testbed.devices[device].connections:
            # if ssh_only is not enabled try to connect through all connections
            if one_connect == 'defaults':
                continue
            if not self.ssh_only:
                try:
                    self.testbed.devices[device].connect(via = str(one_connect),
                                                         connection_timeout=self.timeout,
                                                         log_stdout=to_stdout,
                                                         logfile = self.logfile,
                                                         learn_os = True,
                                                         init_config_commands = self.disable_config)
                    log.debug('     Connected to device {}'.format(device))
                    break
                except Exception as e:
                    log.debug('     Failed to connect to {name} using connection {conn}'.format(name = device, conn = one_connect))                   
                    # if connection fails, erase the connection from connection mgr
                    self.testbed.devices[device].destroy(str(one_connect))
                continue

            # if ssh only is enabled, check if the connection protocol is ssh before trying to connect
            if self.testbed.devices[device].connections[one_connect].get('protocol', '') == 'ssh':
                try:
                    self.testbed.devices[device].connect(via=str(one_connect),
                                                         connection_timeout=self.timeout,
                                                         log_stdout=to_stdout,
                                                         logfile = self.logfile,
                                                         learn_os = True,
                                                         init_config_commands = self.disable_config)
                    log.debug('     Connected to device {}'.format(device))
                    break
                except Exception as e:
                    # if connection fails, erase the connection from connection mgr
                    log.debug('     Failed to connect to {name} using connection {conn}'.format(name = device, conn = one_connect))
                    self.testbed.devices[device].destroy(str(one_connect))
        
        if not self.testbed.devices[device]:
            log.debug('     Failed to connect to {}'.format(device))
        return self.testbed.devices[device].connected
                

    def configure_testbed_cdp_protocol(self):
        ''' Method checks if cdp configuration is necessary for all devices in
        the testbed and if needed calls the cdp configuration method for the
        target devices in parallel
        '''

        # Check which device to configure CDP on
        device_to_configure=[]
        for device_name, device_obj in self.testbed.devices.items():
            if device_name in self.visited_devices or device_name in self.cdp_configured or not device_obj.connected or device_obj.os not in self.supported_os:
                continue
            device_to_configure.append(device_obj)

        # No device to configure
        if not device_to_configure:
            return

        # Configure cdp on these device
        res = pcall(self.configure_device_cdp_protocol,
                    device=device_to_configure)
        for result in res:
            if result[1]:
                self.cdp_configured.add(result[0])        
        

    def configure_device_cdp_protocol(self, device):
        '''If allowed to edit device configuration enable cdp on the device
        Once done - Then add it to the cdp_configured list

        Args:
            device ('device'): the device having cdp enabled
        '''

        # Check if it is already enabled 
        if device.api.verify_cdp_in_state(max_time=self.timeout, check_interval=5):
            # Already configured - Get out
            return(device.name, False)
        
        log.debug('    Configuring cdp protocol for {}'.format(device.name))
        # Configure it
        try:
            device.api.configure_cdp()
        except Exception:
            log.error("     Exception configuring cdp "
                      "for {device}".format(device=device.name),
                                              exc_info=True)
            return(device.name, False)
        else:
            return(device.name, True)

    def configure_testbed_lldp_protocol(self):
        ''' Method checks if lldp configuration is necessary for all devices in
        the testbed and if needed calls the cdp configuration method for the
        target devices in parallel
        '''

        # Check which device to configure lldp on
        device_to_configure = []
        for device_name, device_obj in self.testbed.devices.items():
            if device_name in self.visited_devices or device_name in self.lldp_configured or not device_obj.connected or device_obj.os not in self.supported_os:
                continue
            device_to_configure.append(device_obj)

        # No device to configure    
        if not device_to_configure:
            return

        # Configure lldp on these device    
        res = pcall(self.configure_device_lldp_protocol,
                    device= device_to_configure)
        for result in res:
            if result[1]:
                self.lldp_configured.add(result[0])
        

    def configure_device_lldp_protocol(self, device):
        '''If allowed to edit device configuration enable lldp on the device
        if it is disabled and and then marks that configuration was done

        Args:
            device ('device'): the device having lldp enabled
        '''

        # Check if it is already enabled 
        if device.api.verify_lldp_in_state(max_time= self.timeout, check_interval=5):
            # Already configured - Get out
            return(device.name, False)
        
        log.debug('     Configuring lldp protocol for {}'.format(device.name))
        # Configure it
        try:
            device.api.configure_lldp()
        except Exception:
            log.error("     Exception configuring lldp "
                      "for {device}".format(device=device.name),
                      exc_info=True)
            return(device.name, False)
        else:
            return(device.name, True)

    def get_neigbor_data(self):
        '''Takes a testbed and processes the cdp and lldp data of every
        device on the testbed that has not yet been visited

        Returns:
            [{device:{'cdp':DATA, 'lldp':data}, device2:{'cdp':data,'lldp':data}}]
        '''
        dev_to_test = []
        dev_to_test_names = set()
        # if the device has not been visited add it to the list of devices to test
        # and then add it to list of devices that have been visited
        for device_name, device_obj in self.testbed.devices.items():
            if device_name in self.visited_devices:
                continue
            self.visited_devices.add(device_name)
            if device_obj.os in self.supported_os and device_obj.connected: 
                dev_to_test.append(self.testbed.devices[device_name])
                dev_to_test_names.add(device_name)
            
        # use pcall to get cdp and lldp information for all devices in to test list
        if dev_to_test:
            result = pcall(self.get_neighbor_info, device = dev_to_test)
            return result
        else:
            return []

    def get_neighbor_info(self, device):
        '''Method designed to be used with pcall, gets the devices cdp and lldp
        neighbor data and then returns it in a dictionary format

        Args:
            device ('device'): target to device to call cdp and lldp commands on
        '''
        cdp = {}
        lldp = {}
        if device.os not in self.supported_os or not device.connected:
            return {device.name: {'cdp':cdp, 'lldp':lldp}}

        log.debug('     Getting cdp and lldp neighbor info for {}'.format(device.name))
        
        # get the devices cdp neighbor information
        try:
            cdp = device.api.get_cdp_neighbors_info()
        except Exception as e:
            log.error("     Exception occurred getting cdp info for {}".format(device.name))
            log.debug(e)
        if cdp is None:
            log.debug("     No CDP information found on {}".format(device.name))

        # get the devices lldp neighbor information
        try:
            lldp = device.api.get_lldp_neighbors_info()
        except Exception as e:
            log.error("     Exception occurred getting lldp info for {}".format(device.name))
            log.debug(e)
        if lldp is None:
            log.debug("     No LLDP information found on {}".format(device.name))
        log.debug('     Got cdp and lldp neighbor info for {}'.format(device.name))
        return {device.name: {'cdp':cdp, 'lldp':lldp}}

    def unconfigure_neighbor_discovery_protocols(self, device):
        '''Unconfigures neighbor discovery protocols on device if they
        were enabled by the script earlier

        Args:
            device ('device'): device to unconfigure protocols on
        '''
        log.debug('   Unconfiguring neighbor discovery protocol for {}'.format(device.name))
        # if the device had cdp configured by the script, disable cdp on the device
        if device.name in self.cdp_configured:
            try:
                device.api.unconfigure_cdp()
            except Exception as e:
                log.error('     Error unconfiguring cdp on device {}: {}'.format(device.name, e))

        # if the device had lldp configured by the script, disable lldp on the device
        if device.name in self.lldp_configured:
            try:
                device.api.unconfigure_lldp()
            except Exception as e:
                log.error('     Error unconfiguring lldp on device {}: {}'.format(device.name, e))

    def get_interfaces_ipV4_address(self, device):
        '''Get the ip address for all of the generated interfaces on the give device

        Args:
            device ('device'): device to get interface ip addresses for
        '''
        
        log.debug('   Getting interface ipv4 addresses for {}'.format(device.name))
        # if the device isn't connected or the device doesn't have any interfaces to get ip address for
        if not device.connected or device.os not in self.supported_os or len(device.interfaces) < 1:
            return
        for interface in device.interfaces.values():
            if interface.ipv4 is None:
                try:
                    ip = device.api.get_interface_ipv4_address(interface.name )
                except Exception:
                    ip = None
                if ip:
                    ip = ipaddress.IPv4Interface(ip)
                    interface.ipv4 = ip

    def get_credentials_and_proxies(self, yaml):
        '''Takes a copy of the current credentials in the testbed for use in
        connecting to other devices

        Args:
            yaml ('dict'): testbed to collect credentials and proxies for

        Returns:
            dict of credentials used in connections
            list of proxies used by testbed devices
        '''
        credential_dict = {}
        proxy_list = []
        for device in yaml['devices'].values():
            
            # get all connections used in the testbed
            if 'credentials' in device:
                for cred in device['credentials']:
                    if cred not in credential_dict :
                        credential_dict[cred] = dict(device['credentials'][cred])
                    elif device['credentials'][cred] not in credential_dict.values():
                        credential_dict[cred + str(len(credential_dict))] = dict(device['credentials'][cred])

            # get list of proxies used in connections
            for connect in device['connections'].values():
                if 'proxy' in connect:
                    if connect['proxy'] not in proxy_list:
                        proxy_list.append(connect['proxy'])

        return credential_dict, proxy_list
